library(Matrix)

#################################################################

myentropy <- function (ct) {
    -sum((rowSums(ct) / sum(ct))
          * apply(ct,1,function(pv) {
                         p1 <- pv/sum(pv)
                         p2 <- p1[p1 != 0]
                         sum(p2 * log(p2))
                       })) /log(ncol(ct))
}


mypurity <- function (ct) {
    sum(apply(ct,1,max)) / sum(ct)
}

myeval <- function(myans, ansfile) {
     goldans <- scan(ansfile,what="character")
     ct <- table(myans,goldans)
     cat("Entropy: ",myentropy(ct),"\n")
     cat("Purity : ",mypurity(ct),"\n")
}

#################################################################

tfidf <- function(X) {
     cw <- log(nrow(X) * apply(X,2,function(x) 1/length(x[x > 0])))
     A <- diag(cw)
     return(X %*% A)
}

seikika <- function(X) {
     cw <- apply(X,1,function(x) 1/sqrt(sum(x**2)))
     A <- diag(cw)
     return(A %*% X)
}

#################################################################

myhclust <- function(x, clsn, meth="complete") {
     switch(meth,               
        "ward"     = d <- dist(x)**2,
        "centroid" = d <- dist(x)**2,
        "median"   = d <- dist(x)**2,
        d <- dist(x)
     )  
     hc <- hclust(d, method=meth)
     return(cutree(hc,k=clsn))
}

myhclust2 <- function(infile, clsn) {
     meth <- "ward"
     x1 <- as(readMM(infile),"CsparseMatrix")     
     x1 <- x1 %*% (t(x1))
     x2 <- as.matrix(x1)
     x2 <- 2 * (1 - x2)
     d <- as.dist(x2)
     hc <- hclust(d, method=meth)
     return(cutree(hc,k=clsn))
}

#################################################################

mykmval <- function(x, clsn, ans) {
     N <- nrow(x)
     val <- 0
     for(cl in 1:clsn) {
          g <- c(1:N)[ans == cl]
          cen <- apply(x[g,],2,sum)/length(g)
          val <- val + sum(apply(x[g,],1,function(x) sum((x - cen)**2)))
     }
     return(val)
}


mykmeans2 <- function(infile, clsn, maxc=10) {
     x1 <- as.matrix(readMM(infile))
     N <- nrow(x1)
     mem <- sample(c(1:N),clsn)
     G <- x1[mem,]
     lc <- 0; contp <- TRUE; result0 <- numeric(N) + 1
     while((lc <= maxc) && contp) {
          x2 <- x1 %*% t(G)
          result <- apply(x2,1,which.max)
          if (length(result[result != result0]) == 0) {
               contp <- FALSE
          } else {
               for (i in 1:clsn) {
                    mem <- c(1:N)[result == i]
                    G[i,] <- colSums(x1[mem,])/length(mem)
               }
               result0 <- result
               lc <- lc + 1
          }
     }
     return(result)
}


mykmval2 <- function(X, clsn, ans) {
     N <- nrow(X)
     a <- matrix(c(1:clsn),ncol=1)
     return(sum(apply(a,1,function(x) mykmvalcore(x,N,X,ans))))
}

mykmvalcore <- function(x,N,X,ans) {
     g <- c(1:N)[ans == x]
     cen <- colSums(X[g,])/length(g)
     return(sum(apply(X[g,],1,function(x) sum((x - cen)**2))))
}


#################################################################

mcut <- function (W,clsn) {
     N <- nrow(W)
     gclsn <- 1
     objlist <- c(1)
     result <- numeric(N) + 1
     tgc <- 1
     while(gclsn < clsn) {
          tgv <- c(1:N)[result == tgc]
          cab <- mcutcore(W,tgv)
          ca <- cab[[1]]
          cb <- cab[[2]]
          gclsn <- gclsn + 1          
          result[cb] <- gclsn
          cav1 <- objfv(W,ca)
          cav2 <- objfv(W,cb)
          objlist[tgc] <- cav1
          objlist <- c(objlist,cav2)
          tgc <- which.min(objlist)
     }
     return(result)
}

mcutcore <- function (bigW,tgv) {
     W <- bigW[tgv,tgv]
     cab <- mcutcore2(W)
     ca <- cab[[1]]
     cb <- cab[[2]]
     return(list(tgv[ca],tgv[cb]))
}

mcutcore2 <- function (W) {
     N <- nrow(W)
     D2 <- diag(1/sqrt(apply(W,1,sum)))
     I <- diag(numeric(N) + 1)
     X <- I - D2 %*% W %*% D2
     eg <- eigen(X)
     min1 <- order(eg$values)[2]
     fv <- eg$vectors[,min1]
     ca <- c(1:N)[fv >= 0]
     cb <- c(1:N)[fv < 0]
     return(list(ca,cb))
}

objfv <- function (W,ca) {
     N <- length(ca)
     dzero <- W[1,1]
     if (N == 1) return(dzero)
     return((sum(W[ca,ca]) - N*dzero)/(N*(N-1)))
}

#################################################################

plsi <- function(x,clsn,maxc = 20) {
     d <- initd(x,clsn)
     w <- initw(x,clsn)     
     z <- initz(clsn)
     q <- initq(x,clsn)
     loopc <- 0     
     while(loopc < maxc) {
          q <- newq(q,x,d,w,z,clsn)
          d <- newd(d,x,q,clsn)
          w <- neww(w,x,q,clsn)
          z <- newz(z,x,q,clsn)
          loopc <- loopc + 1
     }
     return(list(d,w,z))
}

newq <- function(q,X,d,w,z,clsn) {
     dn <- nrow(X)
     wn <- ncol(X)
     qall <- matrix(numeric(dn*wn),ncol=wn)
     for(cl in 1:clsn) {
          d0 <- d[,cl]
          w0 <- w[,cl]
          z0 <- z[cl]
          qc <- (d0 %*% t(w0))*z0
          q[,,cl] <- qc
          qall <- qall + qc
     }
     for(cl in 1:clsn) {
          q[,,cl] <- q[,,cl]/qall
     }
     return(q)
}

newd <- function(d,X,q,clsn) {
     for(cl in 1:clsn) {
         Q<- q[,,cl]
         a <- rowSums(X * Q)
         d[,cl] <- a/sum(a)
     }
     return(d)
}

neww <- function(w,X,q,clsn) {
     for(cl in 1:clsn) {
          Q<- q[,,cl]
          a <- colSums(X * Q)
          w[,cl] <- a/sum(a)          
     }
     return(w)
}

newz <- function(z,X,q,clsn) {
     for(cl in 1:clsn) {
          Q<- q[,,cl]
          z[cl] <- sum(X * Q)
     }
     ss <- sum(z)
     return(z/ss)
}

initd <- function(x,clsn) {
     dn <- nrow(x)
     d <- matrix(runif(dn*clsn),ncol=clsn)
     return(d)     
}

initw <- function(x,clsn) {
     wn <- ncol(x)
     w <- matrix(runif(wn*clsn),ncol=clsn)
     return(w)     
}

initz <- function(clsn) {
     z <- numeric(clsn) + 1/clsn
     return(z)
}

initq <- function(x,clsn) {
     dn <- nrow(x)
     wn <- ncol(x)
     q <- array(numeric(dn*wn*clsn),dim=c(dn,wn,clsn))
     return(q)
}

#################################################################

nmf <- function (tX,clsn,maxlc = 30) {
     X <- t(tX)
     n <- nrow(X)
     m <- ncol(X)
     U <- matrix(runif(n*clsn) + 0.1,ncol=clsn)
     U <- normalU(U)
     V <- matrix(runif(m*clsn) + 0.1,ncol=clsn)
     loopc <- 0
     while(loopc < maxlc) {
          v1 <- tX %*% U
          v2 <- V %*% t(U) %*% U
          Vn <- V * v1 / v2
          u1 <- X %*% V
          u2 <- U %*% t(V) %*% V
          Un <- U * u1 / u2
          V <- Vn
          U <- normalU(Un)
          loopc <- loopc + 1
     }
     return(V)
}

normalU <- function(U) {
     NN <- diag(1/apply(U,2,function(x) sqrt(sum(x**2))))
     return(U %*% NN)
}


#################################################################

mycmeans <- function(X,clsn,maxc = 50) {
     n <- nrow(X)
     m <- ncol(X)
     mem <- sample(c(1:n),clsn)
     centers <- X[mem,]
     reltol <- sqrt(.Machine$double.eps)
     weights <- numeric(n) + 1/n
     retval <- .C("cmeans",
                 as.double(X),
                 as.integer(n),
                 as.integer(m),
                 centers = as.double(centers),
                 as.integer(clsn),
                 as.double(weights),
                 as.double(2),
                 as.integer(0),
                 as.integer(maxc),
                 as.double(reltol),
                 as.integer(FALSE),
                 u = double(n * clsn),
                 ermin = double(1),
                 iter = integer(1),
                 PACKAGE = "e1071")
     u <- matrix(retval$u, ncol = clsn)
     ans <- apply(u, 1, which.max)
     return(ans)
}
